"""
Unit tests for decayfunctions.py functions, classes and methods.
"""

import copy
import unittest
import radioactivedecay as rd


class Test(unittest.TestCase):
    """
    Unit tests for decayfunctions.py functions, classes and methods.
    """

    # pylint: disable=too-many-public-methods

    def test_inventory_instantiation(self):
        """
        Test instantiation of Inventory objects.
        """

        inv = rd.Inventory({"H-3": 1.0})
        self.assertEqual(inv.contents, {"H-3": 1.0})
        inv = rd.Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(inv.contents, {"Tc-99m": 2.3, "I-123": 5.8})

    def test_inventory_change(self):
        """
        Test Inventory change() method.
        """

        inv = rd.Inventory({"H-3": 1.0})
        inv.change({"Tc-99m": 2.3, "I-123": 5.8}, True, rd.DEFAULTDATA)
        self.assertEqual(inv.contents, {"Tc-99m": 2.3, "I-123": 5.8})

    def test_inventory_radionuclides(self):
        """
        Test Inventory radionuclides property.
        """

        inv = rd.Inventory({"H-3": 1.0})
        self.assertEqual(inv.radionuclides, ["H-3"])
        inv = rd.Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(inv.radionuclides, ["I-123", "Tc-99m"])

    def test_inventory_activities(self):
        """
        Test Inventory activities property.
        """

        inv = rd.Inventory({"H-3": 1})
        self.assertEqual(inv.activities, [1.0])
        inv = rd.Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(inv.activities, [5.8, 2.3])

    def test_inventory_add(self):
        """
        Test Inventory add() method to append to an inventory.
        """

        inv = rd.Inventory({"H-3": 1})
        inv.add({"C-14": 3.0, "K-40": 4.0})
        self.assertEqual(inv.contents, {"C-14": 3.0, "H-3": 1.0, "K-40": 4.0})
        inv.add({"H-3": 3.0})
        self.assertEqual(inv.contents, {"C-14": 3.0, "H-3": 4.0, "K-40": 4.0})

    def test_inventory_subtract(self):
        """
        Test Inventory subtract() method to take away a dictionary from an inventory.
        """

        inv = rd.Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0})
        inv.subtract({"C-14": 3.0, "K-40": 4.0})
        self.assertEqual(inv.contents, {"C-14": 0.0, "H-3": 4.0, "K-40": 0.0})

    def test_inventory___add__(self):
        """
        Test operator to add two inventory objects together.
        """

        inv1 = rd.Inventory({"H-3": 1.0})
        inv2 = rd.Inventory({"C-14": 1.0, "H-3": 4.0})
        inv = inv1 + inv2
        self.assertEqual(inv.contents, {"C-14": 1.0, "H-3": 5.0})

        temp_data = copy.deepcopy(rd.DEFAULTDATA)
        temp_data.dataset = "icrp107_"
        inv3 = rd.Inventory({"H-3": 2.0}, data=temp_data)
        with self.assertRaises(ValueError):
            inv = inv1 + inv3

    def test_inventory___subtract__(self):
        """
        Test operator to subtract one inventory object from another.
        """

        inv1 = rd.Inventory({"H-3": 1.0})
        inv2 = rd.Inventory({"C-14": 1.0, "H-3": 4.0})
        inv = inv2 - inv1
        self.assertEqual(inv.contents, {"C-14": 1.0, "H-3": 3.0})

        temp_data = copy.deepcopy(rd.DEFAULTDATA)
        temp_data.dataset = "icrp107_"
        inv3 = rd.Inventory({"H-3": 2.0}, data=temp_data)
        with self.assertRaises(ValueError):
            inv = inv1 - inv3

    def test_inventory___mul__(self):
        """
        Test operator to multiply activities in inventory by constant.
        """

        inv = rd.Inventory({"Sr-90": 1.0, "Cs-137": 1.0})
        inv = inv * 2
        self.assertEqual(inv.contents, {"Cs-137": 2.0, "Sr-90": 2.0})

    def test_inventory___rmul__(self):
        """
        Test operator to right multiply constant by activities in inventory.
        """

        inv = rd.Inventory({"Sr-90": 1.0, "Cs-137": 1.0})
        inv = 2 * inv
        self.assertEqual(inv.contents, {"Cs-137": 2.0, "Sr-90": 2.0})

    def test_inventory___truediv__(self):
        """
        Test operator to multiply activities in inventory by constant.
        """

        inv = rd.Inventory({"Sr-90": 1.0, "Cs-137": 1.0})
        inv = inv / 2
        self.assertEqual(inv.contents, {"Cs-137": 0.5, "Sr-90": 0.5})

    def test_inventory_remove(self):
        """
        Test operator to remove radionuclides from an inventory.
        """

        inv = rd.Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0})
        with self.assertRaises(NotImplementedError):
            inv.remove(1.0)

    def test_inventory_remove_string(self):
        """
        Test operator to remove one radionuclide from an inventory.
        """

        inv = rd.Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0})
        inv.remove("H-3")
        self.assertEqual(inv.contents, {"C-14": 3.0, "K-40": 4.0})

        with self.assertRaises(ValueError):
            inv.remove("Be-10")

    def test_inventory_remove_list(self):
        """
        Test operator to remove list of radionuclides from an inventory.
        """

        inv = rd.Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0})
        inv.remove(["H-3", "C-14"])
        self.assertEqual(inv.contents, {"K-40": 4.0})

        with self.assertRaises(ValueError):
            inv.remove(["Be-10", "C-14"])

    def test_inventory_decay(self):
        """
        Test Inventory decay() calculations.
        """

        inv = rd.Inventory({"H-3": 10.0})
        self.assertEqual(inv.decay(12.32, "y").contents, {"H-3": 5.0})
        inv = rd.Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(
            inv.decay(20.0, "h").contents,
            {
                "I-123": 2.040459244534774,
                "Tc-99": 6.729944738772211e-09,
                "Tc-99m": 0.22950748010063513,
                "Te-123": 9.485166535243877e-18,
                "Te-123m": 7.721174031572363e-07,
            },
        )
        inv = rd.Inventory({"U-238": 99.274, "U-235": 0.720, "U-234": 0.005})
        self.assertEqual(
            inv.decay(1e9, "y").contents,
            {
                "Ac-227": 0.2690006281740556,
                "At-218": 0.017002868638497183,
                "At-219": 2.227325201281319e-07,
                "Bi-210": 85.01434361515662,
                "Bi-211": 0.26900084425585846,
                "Bi-214": 85.01432618961896,
                "Bi-215": 2.1605054452429237e-07,
                "Fr-223": 0.0037122086688021884,
                "Hg-206": 1.6152725286830197e-06,
                "Pa-231": 0.2690006198549055,
                "Pa-234": 0.13601313171698984,
                "Pa-234m": 85.00820732310412,
                "Pb-210": 85.01434361489548,
                "Pb-211": 0.2690008442558569,
                "Pb-214": 84.99734032384839,
                "Po-210": 85.01434362236536,
                "Po-211": 0.0007424423301461693,
                "Po-214": 84.99649018398776,
                "Po-215": 0.26900084425583065,
                "Po-218": 85.01434319248591,
                "Ra-223": 0.26900062820528614,
                "Ra-226": 85.01434319228659,
                "Rn-218": 1.7002868638497185e-05,
                "Rn-219": 0.26900062820528614,
                "Rn-222": 85.0143431924858,
                "Th-227": 0.2652884195245263,
                "Th-230": 85.01431274847525,
                "Th-231": 0.26898810215560653,
                "Th-234": 85.00820732310407,
                "Tl-206": 0.00011383420610068998,
                "Tl-207": 0.26825840192571576,
                "Tl-210": 0.01785300849981999,
                "U-234": 85.01287846492669,
                "U-235": 0.2689881021544942,
                "U-238": 85.00820732184867,
            },
        )

    def test_inventory___repr__(self):
        """
        Test Inventory representations.
        """

        inv = rd.Inventory({"H-3": 10.0})
        self.assertEqual(
            inv.__repr__(), "Inventory: {'H-3': 10.0}, Decay dataset: icrp107"
        )

    def test_radionuclide_instantiation(self):
        """
        Test instantiation of Radionuclide objects.
        """

        nuc = rd.Radionuclide("H-3")
        self.assertEqual(nuc.radionuclide, "H-3")

    def test_radionuclide_change(self):
        """
        Test Radionuclide change() method.
        """

        nuc = rd.Radionuclide("H-3")
        nuc.change("Rn-222", rd.DEFAULTDATA)
        self.assertEqual(nuc.radionuclide, "Rn-222")
        self.assertEqual(nuc.decay_constant, 2.0982180755947176e-06)

    def test_radionuclide_half_life(self):
        """
        Test Radionuclide half_life() method.
        """

        nuc = rd.Radionuclide("H-3")
        self.assertEqual(nuc.half_life("y"), 12.32)

    def test_radionuclide_progeny(self):
        """
        Test Radionuclide half_life() method.
        """

        nuc = rd.Radionuclide("K-40")
        self.assertEqual(nuc.progeny()[0], "Ca-40")
        self.assertEqual(nuc.progeny()[1], "Ar-40")

    def test_radionuclide_branching_fractions(self):
        """
        Test Radionuclide branching_fractions() method.
        """

        nuc = rd.Radionuclide("K-40")
        self.assertEqual(nuc.branching_fractions()[0], 0.8914)
        self.assertEqual(nuc.branching_fractions()[1], 0.1086)

    def test_radionuclide_decay_modes(self):
        """
        Test Radionuclide decay_modes() method.
        """

        nuc = rd.Radionuclide("K-40")
        self.assertEqual(nuc.decay_modes()[0], "\u03b2-")
        self.assertEqual(nuc.decay_modes()[1], "\u03b2+ & EC")

    def test_radionuclide___repr__(self):
        """
        Test Radionuclide representations.
        """

        nuc = rd.Radionuclide("H-3")
        self.assertEqual(nuc.__repr__(), "Radionuclide: H-3, Decay dataset: icrp107")


if __name__ == "__main__":
    unittest.main()



"""
The decacydata module defines the ``DecayData`` class. Instances of ``DecayData`` initalize by
reading in dataset files containing radioactive decay data. The instances then store the decay
data, and their methods can be used for basic querying of the decay data.

The examples shown assume the ``radioactivedecay`` package has been imported as:

.. highlight:: python
.. code-block:: python

    >>> import radioactivedecay as rd

"""

import numpy as np
from scipy import sparse
from radioactivedecay.utils import parse_nuclide, parse_radionuclide, time_unit_conv

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources


def get_package_filepath(dataset, filename):
    """
    Returns the path to a decay dataset file when the decay dataset is bundled as a sub-package
    within the ``radioactivedecay`` package.

    Parameters
    ----------
    dataset : str
        Name of the decay dataset.
    filename : str
        Name of the file.

    Returns
    -------
    str
        Path of the decay dataset file.

    """

    with pkg_resources.path(__package__ + "." + dataset, filename) as package_path:
        return package_path


class DecayData:
    """
    Instances of DecayData store a radioactive decay dataset.

    Parameters
    ----------
    dataset : str
        Name of the decay dataset.
    dir_path : str or None, optional
        Path to the directory containing the decay dataset files. Use None if the data are bundled
        as a sub-package of ``radioactivedecay`` (default is None).

    Attributes
    ----------
    dataset : str
        Name of the decay dataset.
    decay_consts : numpy.ndarray
        NumPy array of radionuclide decay constants.
    matrix_c : scipy.sparse.csc.csc_matrix
        SciPy Compressed Sparse Column (CSC) matrix that is used in radioactive decay calculations.
        It is a lower traingular matrix with constant elements that are precalculated from decay
        constants and branching fractions.
    matrix_c_inv : scipy.sparse.csc.csc_matrix
        The inverse of matrix_c.
    matrix_e : scipy.sparse.csc.csc_matrix
        SciPy Compressed Sparse Column (CSC) matrix that is used in radioactive decay calculations.
        It is a diagonal matrix that is pre-allocted here for performance reasons.
    no_radionuclides : int
        Number of radionuclides in the dataset.
    radionuclides : numpy.ndarray
        NumPy array of all radionuclides in the dataset.
    radionuclide_dict : dict
        Dictionary containing radionuclide strings as keys and positions in the radionuclides array
        as values.
    prog_bfs_modes : numpy.ndarray
        NumPy array of dictionaries with first progeny as keys and list with branching fraction and
        decay mode as values.
    year_conv : float
        Conversion factor for number of days in one year.

    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, dataset, dir_path=None):
        self.dataset = dataset
        self.load_data(dir_path)

        self.ln2 = np.log(2)

    def load_data(self, dir_path):
        """
        Reads in radioactive decay dataset files and puts the data into DecayData instance
        attributes.

        Parameters
        ----------
        dir_path : str or None
            Path to the directory containing the decay dataset files, or None if the data are
            bundled as a sub-package of ``radioactivedecay``.

        """

        if dir_path is None:
            data = np.load(
                get_package_filepath(self.dataset, "decay_data.npz"), allow_pickle=True,
            )
        else:
            data = np.load(dir_path + "/decay_data.npz", allow_pickle=True)
        self.radionuclides = data["radionuclides"]
        self.decay_consts = data["decay_consts"]
        self.prog_bfs_modes = data["prog_bfs_modes"]
        self.year_conv = data["year_conv"]

        self.no_radionuclides = self.radionuclides.size
        self.radionuclide_dict = dict(
            zip(self.radionuclides, list(range(0, self.no_radionuclides)))
        )
        self.matrix_e = sparse.csc_matrix(
            (
                np.zeros(self.no_radionuclides),
                (np.arange(self.no_radionuclides), np.arange(self.no_radionuclides)),
            )
        )

        if dir_path is None:
            self.matrix_c = sparse.load_npz(get_package_filepath(self.dataset, "c.npz"))
            self.matrix_c_inv = sparse.load_npz(
                get_package_filepath(self.dataset, "cinverse.npz")
            )
        else:
            self.matrix_c = sparse.load_npz(dir_path + "/c.npz")
            self.matrix_c_inv = sparse.load_npz(dir_path + "/cinverse.npz")

    def half_life(self, radionuclide, units="s"):
        """
        Returns half-life of a radionuclide in chosen units.

        Parameters
        ----------
        radionuclide : str
            Radionuclide string.
        units : str, optional
            Units for half-life (default is seconds). Options are 'ns', 'us', 'ms', 's', 'm', 'h',
            'd', 'y', 'ky', 'My', 'Gy', 'Ty', 'Py', and some of the common spelling variations of
            these time units.

        Returns
        -------
        float
            Radionuclide half-life.

        Examples
        --------
        >>> rd.DEFAULTDATA.half_life('Rn-222')
        3.8235

        """

        conv = (
            1.0
            if units == "s"
            else time_unit_conv(
                1.0, units_from="s", units_to=units, year_conv=self.year_conv
            )
        )

        radionuclide = parse_radionuclide(
            radionuclide, self.radionuclides, self.dataset
        )
        return conv * self.ln2 / self.decay_consts[self.radionuclide_dict[radionuclide]]

    def branching_fraction(self, parent, progeny):
        """
        Returns branching fraction for a parent to progeny decay.

        Parameters
        ----------
        parent : str
            Radionuclide string of parent.
        progeny : str
            Nuclide string of progeny (can be stable or radioactive nuclide).

        Returns
        -------
        float
            Branching fraction (or zero if the progeny argument is not actually a direct progeny of
            the parent).

        Examples
        --------
        >>> rd.DEFAULTDATA.branching_fraction('K-40', 'Ca-40')
        0.8914

        """

        parent = parse_radionuclide(parent, self.radionuclides, self.dataset)
        progeny = parse_nuclide(progeny)
        if progeny in self.prog_bfs_modes[self.radionuclide_dict[parent]]:
            return self.prog_bfs_modes[self.radionuclide_dict[parent]][progeny][0]

        return 0.0

    def decay_mode(self, parent, progeny):
        """
        Returns type of decay mode for a parent to progeny decay.

        Parameters
        ----------
        parent : str
            Radionuclide string of parent.
        progeny : str
            Nuclide string of progeny (can be stable or radioactive nuclide).

        Returns
        -------
        str
            Decay mode type (or '' if the progeny argument is not actually a direct progeny of
            the parent).

        Examples
        --------
        >>> rd.DEFAULTDATA.branching_fraction('K-40', 'Ca-40')
        0.8914

        """

        parent = parse_radionuclide(parent, self.radionuclides, self.dataset)
        progeny = parse_nuclide(progeny)
        if progeny in self.prog_bfs_modes[self.radionuclide_dict[parent]]:
            return self.prog_bfs_modes[self.radionuclide_dict[parent]][progeny][1]

        return ""

    def __repr__(self):
        return "Decay dataset: " + self.dataset



"""
Unit tests for decaydata.py functions, classes and methods.
"""

import unittest
from radioactivedecay import decaydata, icrp107


class Test(unittest.TestCase):
    """
    Unit tests for decaydata.py functions, classes and methods.
    """

    def test_decaydata_instantiation(self):
        """
        Test instantiation of DecayData objects.
        """

        data = decaydata.DecayData("icrp107")
        self.assertEqual(data.dataset, "icrp107")
        self.assertEqual(data.no_radionuclides, 1252)
        self.assertEqual(data.year_conv, 365.2422)
        self.assertEqual(data.radionuclides[0], "Fm-257")
        self.assertEqual(data.radionuclides[-1], "H-3")
        self.assertEqual(data.decay_consts[0], 7.982623693568561e-08)
        self.assertEqual(data.decay_consts[-1], 1.7828715741004621e-09)
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[0], "Cf-253")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][0], 0.9979)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][1], "\u03b1")
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][0], 0.0021)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[-1].keys())[0], "He-3")
        self.assertEqual(list(data.prog_bfs_modes[-1].values())[0][0], 1.0)
        self.assertEqual(list(data.prog_bfs_modes[-1].values())[0][1], "\u03b2-")
        self.assertEqual(data.radionuclide_dict["Fm-257"], 0)
        self.assertEqual(data.radionuclide_dict["H-3"], 1251)

        # check instantiations with supplied dataset path
        data = decaydata.DecayData("icrp107_2", icrp107.__path__[0])
        self.assertEqual(data.dataset, "icrp107_2")
        self.assertEqual(data.no_radionuclides, 1252)
        self.assertEqual(data.year_conv, 365.2422)
        self.assertEqual(data.radionuclides[0], "Fm-257")
        self.assertEqual(data.radionuclides[-1], "H-3")
        self.assertEqual(data.decay_consts[0], 7.982623693568561e-08)
        self.assertEqual(data.decay_consts[-1], 1.7828715741004621e-09)
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[0], "Cf-253")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][0], 0.9979)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[0][1], "\u03b1")
        self.assertEqual(list(data.prog_bfs_modes[0].keys())[1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][0], 0.0021)
        self.assertEqual(list(data.prog_bfs_modes[0].values())[1][1], "SF")
        self.assertEqual(list(data.prog_bfs_modes[-1].keys())[0], "He-3")
        self.assertEqual(list(data.prog_bfs_modes[-1].values())[0][0], 1.0)
        self.assertEqual(list(data.prog_bfs_modes[-1].values())[0][1], "\u03b2-")
        self.assertEqual(data.radionuclide_dict["Fm-257"], 0)
        self.assertEqual(data.radionuclide_dict["H-3"], 1251)

    def test_decaydata_half_life(self):
        """
        Test DecayData half_life() method.
        """

        data = decaydata.DecayData("icrp107")
        self.assertEqual(data.half_life("H-3", "y"), 12.32)

    def test_radionuclide_branching_fraction(self):
        """
        Test DecayData branching_fraction() method.
        """

        data = decaydata.DecayData("icrp107")
        self.assertEqual(data.branching_fraction("K-40", "Ca-40"), 0.8914)
        self.assertEqual(data.branching_fraction("K-40", "H-3"), 0.0)

    def test_radionuclide_decay_mode(self):
        """
        Test DecayData decay_mode() method.
        """

        data = decaydata.DecayData("icrp107")
        self.assertEqual(data.decay_mode("K-40", "Ca-40"), "\u03b2-")
        self.assertEqual(data.decay_mode("K-40", "H-3"), "")

    def test_decaydata___repr__(self):
        """
        Test DecayData representations.
        """

        data = decaydata.DecayData("icrp107")
        self.assertEqual(data.__repr__(), "Decay dataset: icrp107")


if __name__ == "__main__":
    unittest.main()



"""
The utils module contains functions to parse nuclide strings, check the validity of inventory
dictionaries, convert between time units, add two dictionaries together.

The examples shown assume the ``radioactivedecay`` package has been imported as:

.. highlight:: python
.. code-block:: python

    >>> import radioactivedecay as rd

"""

from functools import singledispatch, update_wrapper


def parse_nuclide(nuclide):
    """
    Parses a nuclide string from XXAb or AbXX format to Ab-XX format. Not this function works for
    both radioactive and stable nuclides.

    Parameters
    ----------
    nuclide : str
        Nuclide string.

    Returns
    -------
    str
        Nuclide string parsed in symbol - mass number format.

    Examples
    --------
    >>> rd.utils.parse_nuclide('222Rn')
    'Rn-222'
    >>> rd.utils.parse_nuclide('Ca40')
    'Ca-40'

    """

    letter_flag, number_flag = False, False
    for char in nuclide:
        if char.isalpha():
            letter_flag = True
        if char.isdigit():
            number_flag = True
        if letter_flag and number_flag:
            break

    if not (letter_flag and number_flag) or len(nuclide) < 2 or len(nuclide) > 7:
        raise ValueError(str(nuclide) + " is not a valid nuclide string.")

    while nuclide[0].isdigit():  # Re-order inputs e.g. 99mTc to Tc99m.
        nuclide = nuclide[1:] + nuclide[0]
    if nuclide[0] in ["m", "n"]:
        nuclide = nuclide[1:] + nuclide[0]

    for i in range(1, len(nuclide)):  # Add hyphen e.g. Tc99m to Tc-99m.
        if nuclide[i].isdigit():
            if nuclide[i - 1] != "-":
                nuclide = nuclide[:i] + "-" + nuclide[i:]
            break

    return nuclide


def parse_radionuclide(radionuclide, radionuclides, dataset):
    """
    Parses a radionuclide string and checks whether the radionuclide is contained in the decay
    dataset.

    Parameters
    ----------
    radionuclide : str
        Radionuclide string.
    radionuclides : numpy.ndarray
        NumPy array of all the radionuclides in the decay dataset.
    dataset : str
        Name of the decay dataset.

    Returns
    -------
    str
        Radionuclide string parsed in symbol - mass number format.

    Raises
    ------
    ValueError
        If the radionuclide string is invalid or it is not contained in the decay dataset.

    Examples
    --------
    >>> rd.utils.parse_radionuclide('222Rn', rd.DEFAULTDATA.radionuclides, rd.DEFAULTDATA.dataset)
    'Rn-222'
    >>> rd.utils.parse_radionuclide('Ba137m', rd.DEFAULTDATA.radionuclides, rd.DEFAULTDATA.dataset)
    'Ba-137m'

    """

    original_radionuclide = radionuclide
    radionuclide = parse_nuclide(radionuclide)

    if radionuclide not in radionuclides:
        raise ValueError(
            str(original_radionuclide)
            + " is not a valid radionuclide in "
            + dataset
            + " dataset."
        )

    return radionuclide


def check_dictionary(inv_dict, radionuclides, dataset):
    """
    Checks validity of a dictionary of radionuclides and associated acitivities. Radionuclides
    must be in the decay dataset.

    Parameters
    ----------
    inv_dict : dict
        Dictionary containing radionuclide strings as keys and activities as values.
    radionuclides : numpy.ndarray
        NumPy array of all the radionuclides in the decay dataset.
    dataset : str
        Name of the decay dataset.

    Returns
    -------
    dict
        Dictionary where the contents have been validated and the radionuclide key strings have
        been parsed into symbol - mass number format.

    Raises
    ------
    ValueError
        If an activity key is invalid.

    Examples
    --------
    >>> rd.utils.check_dictionary({'3H': 1.0}, rd.DEFAULTDATA.radionuclides, rd.DEFAULTDATA.dataset)
    {'H-3': 1.0}


    """

    inv_dict = {
        parse_radionuclide(nuc, radionuclides, dataset): act
        for nuc, act in inv_dict.items()
    }
    for nuc, act in inv_dict.items():
        if not isinstance(act, (float, int)):
            raise ValueError(
                str(act) + " is not a valid radioactivity for " + str(nuc) + "."
            )

    return inv_dict


def time_unit_conv(time, units_from, units_to, year_conv):
    """
    Converts a time from one set of units to another.

    Parameters
    ----------
    time : float
        Time before conversion.
    units_from : str
        Time unit before conversion
    units_to : str
        Time unit after conversion
    yeav_conv : float or int
        Conversion factor for number of days in a year.

    Returns
    -------
    float
        Time in new units.

    Raises
    ------
    ValueError
        If one of the time unit parameters is invalid.

    Examples
    --------
    >>> rd.utils.time_unit_conv(1.0, 'd', 'h', rd.DEFAULTDATA.year_conv)
    24.0

    """

    conv = {
        "ns": 1.0e-9,
        "us": 1.0e-6,
        "ms": 1.0e-3,
        "s": 1.0,
        "m": 60.0,
        "h": 3600.0,
        "d": 86400.0,
        "y": 86400.0 * year_conv,
        "sec": 1,
        "second": 1,
        "seconds": 1,
        "hr": 3600.0,
        "hour": 3600.0,
        "hours": 3600.0,
        "day": 86400.0,
        "days": 86400.0,
        "yr": 86400.0 * year_conv,
        "year": 86400.0 * year_conv,
        "years": 86400.0 * year_conv,
        "ky": 86400.0 * year_conv * 1.0e3,
        "My": 86400.0 * year_conv * 1.0e6,
        "Gy": 86400.0 * year_conv * 1.0e9,
        "Ty": 86400.0 * year_conv * 1.0e12,
        "Py": 86400.0 * year_conv * 1.0e15,
    }

    if units_from not in conv:
        raise ValueError(
            str(units_from) + ' is not a valid unit, e.g. "s", "m", "h", "d" or "y".'
        )
    if units_to not in conv:
        raise ValueError(
            str(units_to) + ' is not a valid unit, e.g. "s", "m", "h", "d" or "y".'
        )

    return time * conv[units_from] / conv[units_to]


def add_dictionaries(dict1, dict2):
    """
    Adds together two dictionaries of radionuclies and associated acitivities.

    Parameters
    ----------
    dict1 : dict
        First dictionary containing radionuclide strings as keys and activities as values.
    dict2 : dict
        Second dictionary containing radionuclide strings as keys and activities as values.

    Returns
    -------
    dict
        Combined dictionary containing the radionuclides in both dict1 and dict2, where activities
        have been added together when a radionuclide is present in both input dictionaries.

    Examples
    --------
    >>> dict1 = {'Pm-141': 1.0, 'Rb-78': 2.0}
    >>> dict2 = {'Pm-141': 3.0, 'Rb-90': 4.0}
    >>> rd.utils.add_dictionaries(dict1, dict2)
    {'Pm-141': 4.0, 'Rb-78': 2.0, 'Rb-90': 4.0}

    """

    new_dict = dict1.copy()
    for radionuclide, radioactivity in dict2.items():
        if radionuclide in new_dict:
            new_dict[radionuclide] = new_dict[radionuclide] + radioactivity
        else:
            new_dict[radionuclide] = radioactivity

    return new_dict


def method_dispatch(func):
    """Adds singledispatch support for class methods."""

    dispatcher = singledispatch(func)

    def wrapper(*args, **kw):
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)

    wrapper.register = dispatcher.register
    update_wrapper(wrapper, func)
    return wrapper



"""
``radioactivedecay`` is a Python package for performing radioactive decay
calculations. It supports full decay chains and branching. By default it uses
decay data from ICRP Publication 107. It solves the decay chain differential
equations analytically using NumPy and SciPy linear algebra routines.
"""

__version__ = "0.0.8"

from radioactivedecay.decayfunctions import *



"""
The decacyfunctions module defines the two main classes used by users: the ``Inventory`` and
``Radionuclide`` classes.

The examples shown assume the ``radioactivedecay`` package has been imported as:

.. highlight:: python
.. code-block:: python

    >>> import radioactivedecay as rd

Attributes
----------
DEFAULTDATA : DecayData
    Default decay dataset used by ``radioactivedecay``. This is currently ICRP 107.
"""

import numpy as np
from radioactivedecay.decaydata import DecayData
from radioactivedecay.utils import (
    parse_radionuclide,
    check_dictionary,
    time_unit_conv,
    add_dictionaries,
    method_dispatch,
)

DEFAULTDATA = DecayData("icrp107")


class Inventory:
    """
    Inventory instances store a dictionary of radionuclides and associated activities, and a
    DecayData object of radioactive decay data.

    Parameters
    ----------
    contents : dict
        Dictionary containing radionuclide strings as keys and activities as values.
    check : bool, optional
        Check for the validity of contents (default is True).
    data : DecayData, optional
        Decay dataset (default is the ICRP 107 dataset).

    Attributes
    ----------
    contents : dict
        Dictionary containing radionuclide strings as keys and activities as values. Radionuclides
        in contents are sorted alphabetically.
    data : DecayData
        Decay dataset.

    Examples
    --------
    >>> rd.Inventory({'Tc-99m': 2.3, 'I-123': 5.8})
    Inventory: {'I-123': 5.8, 'Tc-99m': 2.3}, Decay dataset: icrp107

    """

    def __init__(self, contents, check=True, data=DEFAULTDATA):
        self.change(contents, check, data)

    def change(self, contents, check, data):
        """
        Changes or initializes the contents and data attritubes of this Inventory instance.
        """

        if check is True:
            contents = check_dictionary(contents, data.radionuclides, data.dataset)
        self.contents = dict(sorted(contents.items(), key=lambda x: x[0]))
        self.data = data

    @property
    def radionuclides(self):
        """
        Returns a list of the radionuclides in the Inventory.

        Examples
        --------
        >>> rd.Inventory({'Tc-99m': 2.3, 'I-123': 5.8}).radionuclides
        ['I-123', 'Tc-99m']

        """

        return list(self.contents)

    @property
    def activities(self):
        """
        Returns a list of the radionuclides in the Inventory.

        Examples
        --------
        >>> rd.Inventory({'Tc-99m': 2.3, 'I-123': 5.8}).activities
        [5.8, 2.3]

        """

        return list(self.contents.values())

    def add(self, add_contents):
        """
        Adds a dictionary of radionuclides and associated activities to this inventory.

        Parameters
        ----------
        add_contents : dict
            Dictionary containing radionuclide strings as keys and activities as values which are
            added to the Inventory object.

        Examples
        --------
        >>> inv = rd.Inventory({'H-3': 1.0})
        >>> inv.add({'C-14': 2.0})
        >>> inv.contents
        {'C-14': 2.0, 'H-3': 1.0}

        """

        add_contents = check_dictionary(
            add_contents, self.data.radionuclides, self.data.dataset
        )
        new_contents = add_dictionaries(self.contents, add_contents)
        self.change(new_contents, False, self.data)

    def subtract(self, sub_contents):
        """
        Subtracts a dictionary of radionuclides and associated activities from this inventory.

        Parameters
        ----------
        sub_contents : dict
            Dictionary containing radionuclide strings as keys and activities as values which are
            subtracted from the Inventory object.

        Examples
        --------
        >>> inv = rd.Inventory({'C-14': 2.0, 'H-3': 1.0})
        >>> inv.subtract({'H-3': 1.0})
        >>> inv.contents
        {'C-14': 2.0, 'H-3': 0.0}

        """
        sub_contents = check_dictionary(
            sub_contents, self.data.radionuclides, self.data.dataset
        )
        sub_contents.update(
            (nuclide, radioactivity * -1.0)
            for nuclide, radioactivity in sub_contents.items()
        )
        new_contents = add_dictionaries(self.contents, sub_contents)
        self.change(new_contents, False, self.data)

    def __add__(self, other):
        """
        Defines + operator to add two Inventory objects together.
        """

        if self.data.dataset != other.data.dataset:
            raise ValueError(
                "Decay datasets do not match. inv1: "
                + self.data.dataset
                + " inv2: "
                + other.data.dataset
            )
        new_contents = add_dictionaries(self.contents, other.contents)
        return Inventory(new_contents, False, self.data)

    def __sub__(self, other):
        """
        Defines - operator to subtract one Inventory object from another.
        """

        if self.data.dataset != other.data.dataset:
            raise ValueError(
                "Decay datasets do not match. inv1: "
                + self.data.dataset
                + " inv2: "
                + other.data.dataset
            )
        sub_contents = other.contents.copy()
        sub_contents.update(
            (nuclide, radioactivity * -1.0)
            for nuclide, radioactivity in sub_contents.items()
        )
        new_contents = add_dictionaries(self.contents, sub_contents)
        return Inventory(new_contents, False, self.data)

    def __mul__(self, const):
        """
        Defines * operator to multiply all activities of radionuclides in an Inventory by a
        float or int.
        """

        new_contents = self.contents.copy()
        for nuclide, radioactivity in new_contents.items():
            new_contents[nuclide] = radioactivity * const
        return Inventory(new_contents, False, self.data)

    def __rmul__(self, const):
        """
        Defines * operator to multiply all activities of radionuclides in an Inventory by a
        float or int.
        """

        return self.__mul__(const)

    def __truediv__(self, const):
        """
        Defines / operator to divide all activities of radionuclides in an Inventory by a
        float or int.
        """

        return self.__mul__(1.0 / const)

    @method_dispatch
    def remove(self, delete):
        """
        Removes radionuclide(s) from this inventory.

        Parameters
        ----------
        delete : str or list
            Radionuclide string or list of radionuclide strings to delete from the Inventory
            object.

        Examples
        --------
        >>> inv = rd.Inventory({'Be-10': 2.0, 'C-14': 3.0, 'H-3': 1.0, 'K-40': 4.0})
        >>> inv.remove('H-3')
        >>> inv.contents
        {'Be-10': 2.0, 'C-14': 3.0, 'K-40': 4.0}
        >>> inv.remove(['Be-10', 'K-40'])
        {'C-14': 3.0}

        """
        raise NotImplementedError("remove() takes string or list of radionuclides.")

    @remove.register(str)
    def _(self, delete):
        """Remove radionuclide string from this inventory."""
        delete = parse_radionuclide(delete, self.data.radionuclides, self.data.dataset)
        new_contents = self.contents.copy()
        if delete not in new_contents:
            raise ValueError(delete + " does not exist in this inventory.")
        new_contents.pop(delete)
        self.change(new_contents, False, self.data)

    @remove.register(list)
    def _(self, delete):
        """Remove list of radionuclide(s) from this inventory."""
        delete = [
            parse_radionuclide(nuc, self.data.radionuclides, self.data.dataset)
            for nuc in delete
        ]
        new_contents = self.contents.copy()
        for nuc in delete:
            if nuc not in new_contents:
                raise ValueError(nuc + " does not exist in this inventory.")
            new_contents.pop(nuc)
        self.change(new_contents, False, self.data)

    def decay(self, decay_time, units="s"):
        """
        Returns new Inventory resulting from radioactive decay of the current Inventory object for
        decay_time.

        Parameters
        ----------
        decay_time : float
            Decay time.
        units : str, optional
            Units of decay_time (default is seconds). Options are 'ns', 'us', 'ms', 's', 'm', 'h',
            'd', 'y', 'ky', 'My', 'Gy', 'Ty', 'Py', and some of the common spelling variations of
            these time units.

        Returns
        -------
        Inventory
            New Inventory after radioactive decay.

        Examples
        --------
        >>> inv_t0 = rd.Inventory({'H-3': 1.0})
        >>> inv_t1 = inv_t0.decay(12.32, 'y')
        >>> inv_t1.contents
        {'H-3': 0.5}

        """

        decay_time = (
            decay_time
            if units == "s"
            else time_unit_conv(
                decay_time,
                units_from=units,
                units_to="s",
                year_conv=self.data.year_conv,
            )
        )

        vector_n0 = np.zeros([self.data.no_radionuclides], dtype=np.float64)
        indices = set()
        for radionuclide in self.contents:
            i = self.data.radionuclide_dict[radionuclide]
            vector_n0[i] = self.contents[radionuclide] / self.data.decay_consts[i]
            indices.update(self.data.matrix_c[:, i].nonzero()[0])
        indices = list(indices)

        matrix_e = self.data.matrix_e.copy()
        matrix_e.data[indices] = np.exp(
            np.multiply(-decay_time, self.data.decay_consts[indices])
        )
        vector_nt = (
            (self.data.matrix_c.dot(matrix_e)).dot(self.data.matrix_c_inv)
        ).dot(vector_n0)
        vector_at = np.multiply(vector_nt, self.data.decay_consts)

        new_contents = dict(zip(self.data.radionuclides[indices], vector_at[indices]))
        new_contents = dict(sorted(new_contents.items(), key=lambda x: x[0]))
        return Inventory(new_contents, False, self.data)

    def __repr__(self):
        return (
            "Inventory: " + str(self.contents) + ", Decay dataset: " + self.data.dataset
        )


class Radionuclide:
    """
    Radionuclide instances are used to fetch decay data on one radionuclide.

    Parameters
    ----------
    radionuclide : str
        Radionuclide string.
    data : DecayData, optional
        Decay dataset (default is the ICRP 107 dataset).

    Attributes
    ----------
    radionuclide : str
        Radionuclide string.
    decay_constant : numpy.float64
        Decay constant of the radionuclide (s\\ :sup:`-1`).
    prog_bf : dict
        Dictionary containing first progeny as keys and branching fractions as values.
    data : DecayData
        Decay dataset.

    Examples
    --------
    >>> rd.Radionuclide('H-3')
    Radionuclide: H-3, Decay dataset: icrp107

    """

    def __init__(self, radionuclide, data=DEFAULTDATA):
        self.change(radionuclide, data)

    def change(self, radionuclide, data):
        """
        Changes or initializes the radionuclide, decay_constant and data attritubes of this
        Radionuclide instance.
        """

        self.radionuclide = parse_radionuclide(
            radionuclide, data.radionuclides, data.dataset
        )
        self.decay_constant = data.decay_consts[
            data.radionuclide_dict[self.radionuclide]
        ]
        self.prog_bf_mode = data.prog_bfs_modes[
            data.radionuclide_dict[self.radionuclide]
        ]
        self.data = data

    def half_life(self, units="s"):
        """
        Returns half-life of the radionuclide in chosen units.

        Parameters
        ----------
        units : str, optional
            Units for half-life (default is seconds). Options are 'ns', 'us', 'ms', 's', 'm', 'h',
            'd', 'y', 'ky', 'My', 'Gy', 'Ty', 'Py', and some of the common spelling variations of
            these time units.

        Returns
        -------
        float
            Radionuclide half-life.

        Examples
        --------
        >>> Rn222 = rd.Radionuclide('Rn-222')
        >>> Rn222.half_life('d')
        3.8235

        """

        conv = (
            1.0
            if units == "s"
            else time_unit_conv(
                1.0, units_from="s", units_to=units, year_conv=self.data.year_conv
            )
        )
        return conv * self.data.ln2 / self.decay_constant

    def progeny(self):
        """
        Returns the first progeny of the radionuclide.

        Returns
        -------
        list
            List of the first progeny of the radionuclide, ordered by decreasing branching
            fraction.

        Examples
        --------
        >>> K40 = rd.Radionuclide('K-40')
        >>> K40.progeny()
        ['Ca-40', 'Ar-40']

        """

        return list(self.prog_bf_mode.keys())

    def branching_fractions(self):
        """
        Returns the branching fractions for the first progeny of the radionuclide.

        Returns
        -------
        list
            List of branching fractions.

        Examples
        --------
        >>> K40 = rd.Radionuclide('K-40')
        >>> K40.branching_fractions()
        [0.8914, 0.1086]

        """

        return [bf_mode[0] for bf_mode in list(self.prog_bf_mode.values())]

    def decay_modes(self):
        """
        Returns the decay modes creating the first progeny of the radionuclide.

        Returns
        -------
        list
            List of decay modes.

        Examples
        --------
        >>> K40 = rd.Radionuclide('K-40')
        >>> K40.decay_modes()
        ['\u03b2-', '\u03b2+ & EC']

        """

        return [bf_mode[1] for bf_mode in list(self.prog_bf_mode.values())]

    def __repr__(self):
        return (
            "Radionuclide: "
            + str(self.radionuclide)
            + ", Decay dataset: "
            + self.data.dataset
        )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="radioactivedecay",
    version="0.0.8",
    author="Alex Malins",
    author_email="radioactivedecay@REMOVETHISalexmalins.com",
    license="MIT",
    description="A Python package for radioactive decay calculations that supports 1252 radionuclides, including full decay chains and branching.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alexmalins/radioactivedecay",
    project_urls={
        "Bug Tracker": "https://github.com/alexmalins/radioactivedecay/issues",
        "Documentation": "https://alexmalins.com/radioactivedecay",
        "Source Code": "https://github.com/alexmalins/radioactivedecay",
    },
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Topic :: Scientific/Engineering",
        "Topic :: Education ",
    ],
    install_requires=["scipy", "numpy", "importlib_resources; python_version<'3.7'"],
    python_requires=">=3.6",
    packages=setuptools.find_packages(),
    package_data={"": ["icrp107/*.npz"]},
)
